import os
from pathlib import Path

import numpy as np
import torch
from PIL import Image
from qtpy.QtWidgets import QWidget
from sam2.build_sam import build_sam2_video_predictor


# Sam V2 pipeline class
class SamV2_pipeline(QWidget):
    def __init__(
        self,
        napari_viewer,
        main_window_object,
        checkpoint_path,
        model_cfg_name,
    ):
        super().__init__()
        self.viewer = napari_viewer
        self.mwo = main_window_object
        self.source_frame_dir = (
            None  # Will be set inside process volume function
        )

        # torch autocast - following samv2 recommended execution code
        torch.autocast(device_type="cuda", dtype=torch.bfloat16).__enter__()
        if torch.cuda.get_device_properties(0).major >= 8:
            torch.backends.cuda.matmul.allow_tf32 = True
            torch.backends.cudnn.allow_tf32 = True

        DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        sam2_checkpoint = checkpoint_path
        model_cfg = model_cfg_name

        self.predictor = build_sam2_video_predictor(model_cfg, sam2_checkpoint)

        self.preprocess_volume()

        self.inference_state = self.predictor.init_state(
            video_path=self.source_frame_dir.as_posix()
        )

        self.prompts = {}

    def preprocess_volume(self):
        layer_name = self.mwo.image_layers_combo.currentText()
        layer = self.viewer.layers[layer_name]
        volume = layer.data

        # Create a source frame directory
        self.source_frame_dir = Path(self.mwo.interdir_lineedt.text()) / Path(
            layer_name
        )
        print("Creating the frame dir ", self.source_frame_dir)
        self.source_frame_dir.mkdir(parents=True, exist_ok=True)

        # Save each slice as a separate image
        for i in range(volume.shape[0]):
            slice_path = os.path.join(self.source_frame_dir, f"{i:04d}.jpeg")

            if os.path.exists(slice_path):
                continue

            # If the slice path does not exists - create slices and save them
            slice = volume[i, :, :]
            slice_array = Image.fromarray(slice)
            slice_array.save(slice_path)

        print("Frames Generated")

    def add_point(self, point_array, label_id, neg_or_pos=1):
        ann_frame_idx = point_array[0]
        ann_obj_id = label_id
        new_point = [point_array[2], point_array[1]]
        new_label = neg_or_pos
        check_if_our_z_is_new = True
        check_if_our_annotation_is_new = True
        # print(ann_obj_id)

        # Check if in dict else add it
        if ann_obj_id in self.prompts:
            all_list = []
            for existing_list in self.prompts[ann_obj_id]:
                if existing_list[0] == ann_frame_idx:
                    points = existing_list[1]
                    labels = list(existing_list[2])
                    points = np.append(points, [new_point], axis=0)
                    labels.append(new_label)
                    new_list = [
                        ann_frame_idx,
                        points,
                        np.array(labels, np.int32),
                    ]
                    all_list.append(new_list)
                    check_if_our_z_is_new = False
                else:
                    all_list.append(existing_list)

            self.prompts[ann_obj_id] = all_list
            check_if_our_annotation_is_new = False

        else:
            points = np.array(
                [[point_array[2], point_array[1]]], dtype=np.float32
            )
            labels = np.array([neg_or_pos], np.int32)
            self.prompts[ann_obj_id] = [[ann_frame_idx, points, labels]]

        if check_if_our_z_is_new and not (check_if_our_annotation_is_new):
            points = np.array(
                [[point_array[2], point_array[1]]], dtype=np.float32
            )
            labels = np.array([neg_or_pos], np.int32)
            existing_val = self.prompts[ann_obj_id]
            self.prompts[ann_obj_id] = existing_val.append(
                [ann_frame_idx, points, labels]
            )

        # print(labels,points,ann_frame_idx)

        layer_name = self.mwo.output_layers_combo.currentText()
        layer = self.viewer.layers[layer_name]
        label_layer_data = layer.data

        _, out_obj_ids, out_mask_logits = self.predictor.add_new_points(
            inference_state=self.inference_state,
            frame_idx=ann_frame_idx,
            obj_id=ann_obj_id,
            points=points,
            labels=labels,
        )
        mask_for_this_frame = np.zeros(
            (label_layer_data.shape[1], label_layer_data.shape[2]),
            dtype=np.int32,
        )
        for i, out_obj_id in enumerate(out_obj_ids):
            out_mask = (out_mask_logits[i] > 0.0).cpu().numpy()
            # print(out_mask.shape)
            mask_for_this_frame[out_mask[0] == True] = out_obj_id

        label_layer_data[ann_frame_idx, :, :] = mask_for_this_frame
        layer.data = label_layer_data

    def video_propagate(self):
        # run propagation throughout the video and collect the results in a dict
        layer_name = self.mwo.output_layers_combo.currentText()
        layer = self.viewer.layers[layer_name]
        label_layer_data = layer.data
        label_layer_data_2 = label_layer_data.copy()
        firsttime = True

        print("Executing forward")
        for (
            out_frame_idx,
            out_obj_ids,
            out_mask_logits,
        ) in self.predictor.propagate_in_video(
            self.inference_state, start_frame_idx=0
        ):
            mask_for_this_frame = np.zeros(
                (label_layer_data.shape[1], label_layer_data.shape[2]),
                dtype=np.int32,
            )
            for i, out_obj_id in enumerate(out_obj_ids):
                out_mask = (out_mask_logits[i] > 0.0).cpu().numpy()
                # print(out_mask.shape)
                mask_for_this_frame[out_mask[0] == True] = out_obj_id

            label_layer_data[out_frame_idx, :, :] = mask_for_this_frame
            progress = int((out_frame_idx * 50) / label_layer_data.shape[0])
            self.mwo.video_propagation_progressBar.setValue(progress)

        print("Executing reverse")
        for (
            out_frame_idx,
            out_obj_ids,
            out_mask_logits,
        ) in self.predictor.propagate_in_video(
            self.inference_state,
            start_frame_idx=label_layer_data.shape[0] - 1,
            reverse=True,
        ):
            mask_for_this_frame = np.zeros(
                (label_layer_data.shape[1], label_layer_data.shape[2]),
                dtype=np.int32,
            )
            for i, out_obj_id in enumerate(out_obj_ids):
                out_mask = (out_mask_logits[i] > 0.0).cpu().numpy()
                print(out_mask.shape)
                mask_for_this_frame[out_mask[0] == True] = out_obj_id

            label_layer_data_2[out_frame_idx, :, :] = mask_for_this_frame

            if firsttime:
                final_progress_here = int(
                    (out_frame_idx * 50) / label_layer_data.shape[0]
                )
                firsttime = False

            progress = int((out_frame_idx * 50) / label_layer_data.shape[0])
            progress = abs(final_progress_here - progress) + 50
            self.mwo.video_propagation_progressBar.setValue(progress)

        layer.data = np.maximum(label_layer_data, label_layer_data_2)
        # self.mwo.video_propagation_progressBar.setValue(progress)
        self.mwo.video_propagation_progressBar.setValue(100)

    def reset(self):
        self.predictor.reset_state(self.inference_state)
        layer_name = self.mwo.output_layers_combo.currentText()
        layer = self.viewer.layers[layer_name]
        label_layer_data = layer.data
        zero_mask = np.zeros(label_layer_data.shape, dtype=np.int32)
        layer.data = zero_mask



# Imports
import os

import napari
import requests
from pipelines.samv2.Samv2_pipeline_handler import SamV2_pipeline
from qtpy import uic
from qtpy.QtWidgets import (
    QComboBox,
    QFileDialog,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QWidget,
)


# Main Plugin class that is connected from outside at napari plugin entry point
class SAMV2_min(QWidget):
    def __init__(self, napari_viewer):
        # Initializing
        super().__init__()
        self.viewer = napari_viewer

        # Load the UI file - Main window
        script_dir = os.path.dirname(__file__)
        ui_file_name = "SAM_V2.ui"
        abs_file_path = os.path.join(
            script_dir, "..", "UI_files", ui_file_name
        )
        uic.loadUi(abs_file_path, self)

        # Get required children for functionality addition
        self.image_layers_combo = self.findChild(
            QComboBox, "Image_layer_combo"
        )
        self.interdir_lineedt = self.findChild(QLineEdit, "inter_lineedt")
        self.interdir_browse_btn = self.findChild(
            QPushButton, "inter_browse_btn"
        )
        self.output_layers_combo = self.findChild(
            QComboBox, "output_layer_combo"
        )
        self.model_cbbox = self.findChild(QComboBox, "model_cbbox")

        self.initialize_btn = self.findChild(QPushButton, "Initialize_btn")
        self.video_propagation_progressBar = self.findChild(
            QProgressBar, "Propagation_progress"
        )
        self.video_propagate_btn = self.findChild(QPushButton, "Propagate_btn")
        self.reset_btn = self.findChild(QPushButton, "reset_btn")

        # Populate combo box - call
        self.populate_combo_box(self.image_layers_combo, "image")
        self.populate_combo_box(self.output_layers_combo, "label")
        self.populate_model_combo()

        # Connect events to functions
        self.viewer.layers.events.inserted.connect(self.layer_changed)
        self.viewer.layers.events.removed.connect(self.layer_changed)
        self.viewer.layers.events.changed.connect(self.layer_changed)
        self.viewer.mouse_drag_callbacks.append(self.on_mouse_click)

        # Connect button to functions
        self.interdir_browse_btn.clicked.connect(self.choose_inter_frame_dir)
        self.initialize_btn.clicked.connect(self.initialize_pipeline)
        self.video_propagate_btn.clicked.connect(self.video_propagate)
        self.reset_btn.clicked.connect(self.reset_everything)

    # Function to populate combo boxes based on layers
    def populate_combo_box(self, combobx, layer_type="image"):
        # Clear the combo box first
        combobx.clear()
        if layer_type == "image":
            # Get all existing image layers from the napari viewer
            layers = [
                layer.name
                for layer in self.viewer.layers
                if isinstance(layer, napari.layers.Image)
            ]
        elif layer_type == "label":
            # Get all existing label layers from the napari viewer
            layers = [
                layer.name
                for layer in self.viewer.layers
                if isinstance(layer, napari.layers.Labels)
            ]
        else:
            raise ValueError(
                "Invalid layer_type. Expected 'image' or 'label'."
            )

        combobx.addItems(layers)

    # Function to handle combobox state change
    def layer_changed(self):
        # Populate combo box - call
        self.populate_combo_box(self.image_layers_combo, "image")
        self.populate_combo_box(self.output_layers_combo, "label")

    # Choose inter frame dir
    def populate_model_combo(self):
        self.model_cbbox.clear()
        self.model_cbbox.addItems(
            ["sam2_hiera_large", "sam2_hiera_small", "sam2_hiera_tiny"]
        )

    # Choose inter frame dir
    def choose_inter_frame_dir(self):
        dname = QFileDialog.getExistingDirectory()
        print(dname)
        self.interdir_lineedt.setText(str(dname))

    # Initialize pipeline
    BASE_URL = "https://dl.fbaipublicfiles.com/segment_anything_2/072824/"
    CHECKPOINTS = {
        "sam2_hiera_tiny": "sam2_hiera_tiny.pt",
        "sam2_hiera_small": "sam2_hiera_small.pt",
        "sam2_hiera_base_plus": "sam2_hiera_base_plus.pt",
        "sam2_hiera_large": "sam2_hiera_large.pt",
    }

    def initialize_pipeline(self):
        script_dir = os.path.dirname(__file__)

        model_map = {
            "sam2_hiera_large": ("sam2_hiera_l.yaml", "sam2_hiera_large.pt"),
            "sam2_hiera_small": ("sam2_hiera_s.yaml", "sam2_hiera_small.pt"),
            "sam2_hiera_tiny": ("sam2_hiera_t.yaml", "sam2_hiera_tiny.pt"),
        }

        selected_model = self.model_cbbox.currentText()

        if selected_model in model_map:
            model_cfg, checkpoint_name = model_map[selected_model]
            checkpoint_path = os.path.join(
                script_dir, "..", "model", checkpoint_name
            )

            # Check if the checkpoint file exists
            if not os.path.exists(checkpoint_path):
                print(
                    f"Checkpoint {checkpoint_name} not found. Downloading..."
                )
                self.download_checkpoint(checkpoint_name, checkpoint_path)

            self.pipeline_object = SamV2_pipeline(
                self.viewer, self, checkpoint_path, model_cfg
            )
        else:
            print("Model not recognized.")

    def download_checkpoint(self, checkpoint_name, checkpoint_path):
        url = self.BASE_URL + checkpoint_name

        try:
            response = requests.get(url, stream=True)
            response.raise_for_status()  # Check if the download was successful

            with open(checkpoint_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
            print(f"{checkpoint_name} downloaded successfully.")

        except requests.exceptions.RequestException as e:
            print(
                f"Failed to download {checkpoint_name} from {url}. Error: {e}"
            )

    def on_mouse_click(self, layer, event):
        # Check if it is a middle mouse click event
        if event.button == 3:  # 3 represents the middle mouse button
            if "Control" in event.modifiers:
                # print(f'Ctrl + Middle mouse click at {event.position}')
                point = [
                    int(event.position[0]),
                    int(event.position[1]),
                    int(event.position[2]),
                ]
                layer_name = self.output_layers_combo.currentText()
                layer = self.viewer.layers[layer_name]
                active_label = layer.selected_label
                # Negative point
                self.pipeline_object.add_point(
                    point, active_label, neg_or_pos=0
                )

            else:
                # print(f'Middle mouse click at {event.position}')
                point = [
                    int(event.position[0]),
                    int(event.position[1]),
                    int(event.position[2]),
                ]
                layer_name = self.output_layers_combo.currentText()
                layer = self.viewer.layers[layer_name]
                active_label = layer.selected_label
                # positive point
                self.pipeline_object.add_point(
                    point, active_label, neg_or_pos=1
                )

    def video_propagate(self):
        self.pipeline_object.video_propagate()

    def reset_everything(self):
        self.pipeline_object.reset()
